
# THIS FILE IS GENERATED FROM THEANO SETUP.PY
short_version = '0.6.0rc3'
version = '0.6.0rc3'
git_revision = 'c55cb65c146ebea89b93dd7a6e587c155bdd3045'
full_version = '0.6.0rc3.dev-%(git_revision)s' % {
        'git_revision': git_revision}
release = True

if not release:
    version = full_version
